from . import _config
from ._logger import DeanerfiLogger
from typing import Any


def get_logger() -> DeanerfiLogger:
    return _config.LOGGER


def set_logger(logger: DeanerfiLogger) -> None:
    _config.LOGGER = logger


def info(name: str, log_str: str, data: Any = None) -> None:
    _config.LOGGER.info(name=name, log_str=log_str, data=data)


def debug(name: str, log_str: str, data: Any = None) -> None:
    _config.LOGGER.debug(name=name, log_str=log_str, data=data)


def warn(name: str, log_str: str, data: Any = None) -> None:
    _config.LOGGER.warn(name=name, log_str=log_str, data=data)


def critical(name: str, log_str: str, data: Any = None) -> None:
    _config.LOGGER.critical(name=name, log_str=log_str, data=data)


def error(name: str, log_str: str, data: Any = None) -> None:
    _config.LOGGER.error(name=name, log_str=log_str, data=data)



from typing import List, Set

from eth_typing import ChecksumAddress

from deanerfi.utils.data import load_json, write_json
from deanerfi.utils.mongodb import MongoWrapper
from ._adaptors import DexRepo
from ._dtypes import DexInfo, DexPoolInfo


class LocalDexRepo(DexRepo):

    def __init__(self, file_path: str = None, save_to_file: bool = True) -> None:
        self.file_path = file_path
        self.save_to_file = save_to_file
        self._configs: List[DexInfo] = list()
        self._pools: Set[DexPoolInfo] = set()
        if self.file_path:
            self._load()

    def get_configs(self, subtype: str = None, chain_id: int = None) -> List[DexInfo]:
        return_configs = self._configs
        if subtype:
            return_configs = [config for config in return_configs if config.subtype == subtype]
        if chain_id:
            return_configs = [config for config in return_configs if config.chain_id == chain_id]
        return return_configs

    def get_config(self, dex_id: str = None) -> DexInfo:
        return next(dex_config for dex_config in self._configs if dex_config.id == dex_id)

    def delete_dex(self, dex_id: str) -> None:
        self._configs = [dex_config for dex_config in self._configs if dex_config.id != dex_id]
        self._pools = [pool for pool in self._pools if pool.dex_id != dex_id]
        self._save_file()

    def token_pools(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        return [pool for pool in self.chain_pools(chain_id=chain_id) if pool.contains_all(tokens)]

    def pools_w_tokens(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        return [pool for pool in self.chain_pools(chain_id=chain_id) if pool.contains_any(tokens)]

    def dex_pools(self, dex_id: str) -> List[DexPoolInfo]:
        return [pool for pool in self._pools if pool.dex_id == dex_id]

    def get_pool_info(self, chain_id: int, address: ChecksumAddress) -> DexPoolInfo:
        return next(
            pool_info for pool_info in self._pools if
            pool_info.chain_id == chain_id and pool_info.address == address
        )

    def chain_pools(self, chain_id: int) -> List[DexPoolInfo]:
        return [pool for pool in self._pools if pool.chain_id == chain_id]

    def save_pools(self, pools: List[DexPoolInfo]) -> None:
        self._pools.update(set(pools))
        self._save_file()

    def save_configs(self, configs: List[DexInfo]) -> None:
        raise NotImplementedError

    def _load(self) -> None:
        dex_data = load_json(self.file_path)
        configs = dex_data.get('configs')
        pools = dex_data.get('pools')
        if configs:
            self._configs = [self._config_from_dict(conf) for conf in configs]
        if pools:
            self._pools = {self._pool_from_dict(pool) for pool in pools}

    def _save_file(self) -> None:
        if self.file_path and self.save_to_file:
            write_json(
                file_path=self.file_path,
                json_data={
                    'configs': [config.to_dict() for config in self._configs],
                    'pools': [pool.to_dict() for pool in self._pools]
                }
            )


class MongoDBDexRepo(DexRepo):

    def __init__(self, mongo_wrapper: MongoWrapper = None) -> None:
        self._mongo = mongo_wrapper or MongoWrapper()

    def get_configs(self, subtype: str = None, chain_id: int = None) -> List[DexInfo]:
        queue = {}
        if subtype:
            queue['subtype'] = subtype
        if chain_id:
            queue['chain_id'] = chain_id
        response = self._mongo.get('dex_configs', queue)
        return [self._config_from_dict(config) for config in response]

    def get_config(self, dex_id: str = None) -> DexInfo:
        response = self._mongo.get('dex_configs', {'id': dex_id})
        return self._config_from_dict(response[0])

    def delete_dex(self, dex_id: str) -> None:
        self._mongo.delete_one('dex_configs', {'id': dex_id})
        self._mongo.delete_many('dex_pools', {'dex_id': dex_id})

    def token_pools(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        q_filter = {
            'chain_id': chain_id,
            'tokens': {
                '$all': tokens
            }
        }
        response = self._mongo.get('dex_pools', q_filter)
        return [self._pool_from_dict(pool) for pool in response]

    def pools_w_tokens(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        q_filter = {
            'chain_id': chain_id,
            '$or': [
                {'tokens': {'$in': [token]}} for token in tokens
            ]
        }
        response = self._mongo.get('dex_pools', q_filter)
        return [self._pool_from_dict(pool) for pool in response]

    def dex_pools(self, dex_id: str) -> List[DexPoolInfo]:
        response = self._mongo.get('dex_pools', {'dex_id': dex_id})
        return [self._pool_from_dict(pool) for pool in response]

    def get_pool_info(self, chain_id: int, address: ChecksumAddress) -> DexPoolInfo:
        response = self._mongo.get(
            'dex_pools',
            {'chain_id': chain_id, 'address': address}
        )
        return self._pool_from_dict(response[0])

    def chain_pools(self, chain_id: int) -> List[DexPoolInfo]:
        response = self._mongo.get('dex_pools', {'chain_id': chain_id})
        return [self._pool_from_dict(pool) for pool in response]

    def save_pools(self, pools: List[DexPoolInfo]) -> None:
        for pool in pools:
            self._mongo.upsert(
                table='dex_pools',
                data=pool.to_dict(),
                id_dict={
                    'address': pool.address,
                    'dex_id': pool.dex_id,
                    'chain_id': pool.chain_id
                }
            )

    def save_configs(self, configs: List[DexInfo]) -> None:
        for dex_info in configs:
            self._mongo.upsert(
                table='dex_configs',
                data=dex_info.to_dict(),
                id_dict={
                    'dex_id': dex_info.id
                }
            )



from typing import Protocol, Dict, Tuple

from eth_typing import ChecksumAddress

from deanerfi.utils.w3 import cs_addr
from deanerfi.w3 import ContractInfo
from deanerfi.utils.data import load_json, write_json, dataclass_from_dict
from deanerfi.utils.mongodb import MongoWrapper


class ContractRepo(Protocol):

    def get(self, chain_id: int, address: str) -> ContractInfo:
        pass

    def save(self, contract_info: ContractInfo) -> None:
        pass


class LocalContractRepo:
    """In memory Chain Repo that can be loaded/ saved to a json file
    Default file location is '<LOCAL_REPO_PATH>/token-repo.json'
    """

    def __init__(self, file_path: str = None, save_to_file: bool = True) -> None:
        self.file_path = file_path
        self.save_to_file = save_to_file
        self._contracts: Dict[Tuple[int, ChecksumAddress], ContractInfo] = {}
        if self.file_path:
            self._load()

    def _load(self) -> None:
        contract_dict = load_json(self.file_path)
        for contract in contract_dict:
            c_info = ContractInfo(**contract)
            self._contracts[(c_info.chain_id, c_info.address)] = c_info

    def get(self, chain_id: int, address: str) -> ContractInfo:
        return self._contracts.get((chain_id, cs_addr(address)))

    def save(self, contract_info: ContractInfo) -> None:
        self._contracts[(contract_info.chain_id, contract_info.address)] = contract_info
        if self.save_to_file and self.file_path:
            write_json(self.file_path, [contract.to_dict() for contract in self._contracts.values()])


class MongoDBContractRepo:
    _table: str = 'contracts'

    def __init__(self, mongo_wrapper: MongoWrapper = None) -> None:
        self._mongo = mongo_wrapper or MongoWrapper()

    def get(self, chain_id: int, address: str) -> ContractInfo:
        q = {'chain_id': chain_id, 'address': address}
        response = self._mongo.get(self._table, q)[0]
        return dataclass_from_dict(data_dict=response, d_class=ContractInfo)

    def save(self, contract_info: ContractInfo) -> None:
        self._mongo.upsert(
            self._table,
            contract_info.to_dict(),
            {'chain_id': contract_info.chain_id, 'address': contract_info.address}
        )



import asyncio
from dataclasses import dataclass, field
from typing import List
from eth_typing import ChecksumAddress

from deanerfi.dex import DexPoolAPI
from deanerfi.common.msgbus import MessageBus, Event, Command, Message
from decimal import Decimal


@dataclass
class SwapEvent(Event):
    dex_id: str
    chain_id: int
    pool_address: ChecksumAddress
    name: str = field(init=False, default='swap_event')


@dataclass
class UpdateReserves(Command):
    dex_id: str
    chain_id: int
    pool_address: ChecksumAddress
    name: str = field(init=False, default='update_reserves')


@dataclass
class ReservesUpdated(Event):
    dex_id: str
    chain_id: int
    pool_address: ChecksumAddress
    token0: str
    token1: str
    token0_res: Decimal
    token1_res: Decimal
    price: Decimal = None
    name: str = field(init=False, default='update_reserves')

    def __post_init__(self):
        self.price = self.token0_res / self.token1_res


async def listen_for_swaps(pool_api: DexPoolAPI, msg_bus: MessageBus):
    event_filter = pool_api._contract.events.Swap.createFilter(fromBlock='latest')
    while True:
        for swap in event_filter.get_new_entries():
            msg_bus.send_msg(SwapEvent(
                dex_id=pool_api.dex_id, chain_id=pool_api.chain_id, pool_address=pool_api.address
            ))
            msg_bus.send_msg(UpdateReserves(
                dex_id=pool_api.dex_id, chain_id=pool_api.chain_id, pool_address=pool_api.address
            ))
        msg_bus.process_msgs()
        await asyncio.sleep(0.001)


def update_reserves(cmd: UpdateReserves) -> List[Message]:
    dex_repo = deanerfi.dex.api.get_dex_repo()
    pool_info = dex_repo.get_pool_info(chain_id=cmd.chain_id, address=cmd.pool_address)
    pool_api = deanerfi.dex.api.get_pool_api(pool_info)
    pool_reserves = pool_api.get_reserves()
    token_repo = deanerfi.w3.api.get_token_repo()
    tokens = [token_repo.get_token(chain_id=cmd.chain_id, address=addr) for addr in pool_info.tokens]
    return [
        ReservesUpdated(
            dex_id=cmd.dex_id, chain_id=cmd.chain_id, pool_address=cmd.pool_address, token0=tokens[0].symbol,
            token1=tokens[1].symbol, token0_res=tokens[0].from_wei(pool_reserves[tokens[0].address]),
            token1_res=tokens[1].from_wei(pool_reserves[tokens[1].address])
        )
    ]


def pub_dex_trades(chain_id: int, tokens: List[ChecksumAddress], msg_bus: MessageBus) -> None:
    dex_repo = deanerfi.dex.api.get_dex_repo()

    pools_info = dex_repo.pools_w_tokens(chain_id=chain_id, tokens=tokens)
    watch_pools = [deanerfi.dex.api.get_pool_api(pool_info=p_info) for p_info in pools_info]
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            asyncio.gather(
                *[listen_for_swaps(pool, msg_bus) for pool in watch_pools]
            )
        )
    finally:
        loop.close()



from deanerfi import w3, dex


def mongo_repos():
    """Sets up the mongo repositories for the token, chain, and dex repositories."""
    w3.api.set_token_repo(w3.MongoDBTokenRepo())
    w3.api.set_chain_repo(w3.MongoDBChainRepo())
    dex.api.set_dex_repo(dex.MongoDBDexRepo())



from pymongo import MongoClient
from pymongo import database as mongo_database
from os import environ


class MongoWrapper:
    _client: MongoClient
    db: mongo_database

    def __init__(self, host: str = None, port: int = None, user: str = None, password: str = None,
                 database: str = None) -> None:
        self.host = host or environ.get('MONGODB_HOST')
        self.port = port or int(environ.get('MONGODB_PORT'))
        self.user = user or environ.get('MONGODB_USER')
        self.password = password or environ.get('MONGODB_PASSWORD')
        self.database = database or environ.get('MONGODB_DB')
        self._connect()

    def _connect(self) -> None:
        self._client = MongoClient(
            host=self.host, port=self.port, username=self.user, password=self.password
        )
        self.db = self._client[self.database]

    def get(self, table: str, query: dict = None):
        return self.db[table].find(query)

    def get_one(self, table: str, query: dict) -> dict:
        return self.db[table].find_one(filter=query)

    def put(self, table: str, data: dict):
        self.db[table].insert_one(data)

    def update(self, table: str, update_data: dict, id_dict: dict):
        self.db[table].update_one(
            id_dict,
            {'$set': update_data}
        )

    def upsert(self, table: str, data: dict, id_dict: dict):
        self.db[table].update_one(id_dict, {'$set': data}, upsert=True)

    def delete_one(self, table: str, id_dict: dict) -> None:
        self.db[table].delete_one(id_dict)

    def delete_many(self, table: str, id_dict: dict) -> None:
        self.db[table].delete_many(id_dict)


def setup_db(mongo: MongoWrapper) -> None:
    collects = ['dex_pools', 'tokens', 'dex_configs']
    dex_pools = mongo.db['dex_pools']
    dex_pools.create_index([('address', 1), ('dex', 1)], unique=True)



import importlib.resources as pkg_resources
from typing import Type, Union

from deanerfi import data
import json
from dataclasses import dataclass


def _load_pkg_data(file: str):
    return json.loads(
        pkg_resources.read_text(data, file)
    )


def erc20_abi() -> list:
    return _load_pkg_data('erc20-abi.json')


def load_json(file_path: str):
    with open(file_path) as f:
        json_data = json.load(f)
    return json_data


def write_json(file_path: str, json_data: Union[list, dict]) -> None:
    with open(file_path, 'w') as f:
        json.dump(json_data, f)


def dataclass_from_dict(data_dict: dict, d_class: Type[dataclass]):
    return d_class(**{
        k: data_dict.get(k) for k, f in d_class.__dataclass_fields__.items() if f.init
    })



from ._contracts import ContractRepo, LocalContractRepo, MongoDBContractRepo
from ._chains import ChainRepo, LocalChainRepo, MongoDBChainRepo
from ._tokens import TokenRepo, LocalTokenRepo, MongoDBTokenRepo



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any
from time import gmtime
from enum import Enum

from deanerfi.utils.mongodb import MongoWrapper


class LogLevel(Enum):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARN = logging.WARNING
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR


class DeanerfiLogger(ABC):

    @abstractmethod
    def log(self, name: str, lvl: LogLevel, log_str: str, data: Any = None) -> None:
        raise NotImplementedError

    def info(self, name: str, log_str: str, data: Any = None) -> None:
        self.log(name, LogLevel.INFO, log_str, data=data)

    def debug(self, name: str, log_str: str, data: Any = None) -> None:
        self.log(name, LogLevel.DEBUG, log_str, data=data)

    def warn(self, name: str, log_str: str, data: Any = None) -> None:
        self.log(name, LogLevel.WARN, log_str, data=data)

    def critical(self, name: str, log_str: str, data: Any = None) -> None:
        self.log(name, LogLevel.CRITICAL, log_str, data=data)

    def error(self, name: str, log_str: str, data: Any = None) -> None:
        self.log(name, LogLevel.ERROR, log_str, data=data)


class LocalLogger(DeanerfiLogger):

    def __init__(self) -> None:
        logging.Formatter.converter = gmtime
        logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s : %(name)s %(message)s')

    @staticmethod
    def get_logger(name: str) -> logging.Logger:
        return logging.getLogger(name)

    def log(self, name: str, lvl: LogLevel, log_str: str, data: Any = None) -> None:
        _logger = self.get_logger(name)
        _logger.log(level=lvl.value, msg=log_str)
        if data:
            _logger.log(level=lvl.value, msg=data)


class MongoDBLogger(DeanerfiLogger):
    _table: str = 'logs'

    def __init__(self, mongo_wrapper: MongoWrapper = None) -> None:
        self._mongo = mongo_wrapper or MongoWrapper()

    def log(self, name: str, lvl: LogLevel, log_str: str, data: Any = None) -> None:
        self._mongo.put(
            self._table,
            {
                'datetime': str(datetime.utcnow()),
                'name': name,
                'log_level': lvl.value,
                'msg': log_str,
                'data': data
            }
        )



from deanerfi import w3, dex
from deanerfi.arb import TokenGraphTradeRouter

chain_id = 250

# dex.api.set_dex_repo(dex.MongoDBDexRepo())
# w3.api.set_token_repo(w3.MongoDBTokenRepo())
#
# dex_repo = dex.api.get_dex_repo()
# pool_infos = dex_repo.chain_pools(chain_id)
#
# router = TokenGraphTradeRouter(
#     pools=pool_infos
# )
#
# token0 = w3.api.get_token(address='0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83', chain_id=chain_id)
# WAVAX = w3.api.get_token(address='0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7', chain_id=chain_id)

# routes = router.route(
#     token0=USDC.address,
#     token1=WAVAX.address,
#     max_swaps=1
# )
#
# routes = router.loop_route(
#     token0=token0.address,
#     max_swaps=3
# )
#
# for swap_route in routes:
#     swap_route.route[0].amount_in = token0.to_wei(100)
#     dex.api.quote_route(swap_route, chain_id)
#     dex.api.print_swap_route(swap_route, chain_id=chain_id)






from dataclasses import dataclass, field, asdict
from decimal import Decimal
from typing import List, Dict

from eth_typing import ChecksumAddress
from web3.constants import ADDRESS_ZERO

from deanerfi.utils.w3 import cs_addr
from deanerfi.w3 import ContractInfo


@dataclass(kw_only=True)
class DexInfo:
    name: str
    subtype: str
    chain_id: int
    id: str
    trading_fee: Decimal
    contracts: Dict[str, ContractInfo]
    pool_abi: List = None
    subgraph_url: str = None

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass(kw_only=True)
class DexPoolInfo:
    address: ChecksumAddress
    tokens: List[ChecksumAddress]
    dex_id: str
    chain_id: int

    def __post_init__(self) -> None:
        self.tokens = [cs_addr(addr) for addr in self.tokens]
        self.tokens.sort()
        self.address = cs_addr(self.address)

    def __eq__(self, other: "DexPoolInfo") -> bool:
        return self.tokens == other.tokens

    def __str__(self) -> str:
        return f'{self.dex_id}: ' + ' - '.join([str(t) for t in self.tokens])

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self):
        return hash(self.address)

    def to_dict(self) -> dict:
        return asdict(self)

    def contains_any(self, tokens: List[ChecksumAddress]) -> bool:
        """Returns true if the pool contains any of the tokens in the list"""
        return any(token in self.tokens for token in tokens)

    def contains_all(self, tokens: List[ChecksumAddress]) -> bool:
        """Returns true if the pool contains all of the tokens in the list"""
        return all(token in self.tokens for token in tokens)

    @staticmethod
    def null_pool(tokens: List[ChecksumAddress]) -> 'DexPoolInfo':
        tokens.sort()
        return DexPoolInfo(
            address=ADDRESS_ZERO,
            tokens=tokens,
            dex_id='NULL',
            chain_id=-1
        )


@dataclass(kw_only=True)
class Swap:
    token_in: ChecksumAddress
    token_out: ChecksumAddress
    pool_address: ChecksumAddress
    amount_in: Decimal = field(default=Decimal('0'))
    amount_out: Decimal = field(default=Decimal('0'))

    def __eq__(self, other: 'Swap') -> bool:
        return self.pool_address == other.pool_address

    def __str__(self) -> str:
        return f'Pool: {self.pool_address}\n\t{self.amount_in} {self.token_in} -> {self.amount_out} {self.token_out}'

    def __repr__(self) -> str:
        return str(self)


@dataclass(kw_only=True)
class SwapRoute:
    route: List[Swap] = field(default_factory=list)

    class SwapRouteError(Exception):
        pass

    def add_swap(self, next_swap: Swap) -> None:
        if not self.route:
            self.route.append(next_swap)
            return
        _invalid_amount_in = next_swap.amount_in != self.route[-1].amount_out
        _invalid_token_in = next_swap.token_in != self.route[-1].token_out
        if _invalid_token_in or _invalid_amount_in:
            raise self.SwapRouteError(f'Swap Route Error: add_swap ')
        self.route.append(next_swap)

    @property
    def amount_in(self) -> Decimal:
        return self.route[0].amount_in

    @property
    def amount_out(self) -> Decimal:
        return self.route[-1].amount_out

    @property
    def token_in(self) -> ChecksumAddress:
        return self.route[0].token_in

    @property
    def token_out(self) -> ChecksumAddress:
        return self.route[-1].token_out

    def __str__(self) -> str:
        return f'Route: {self.amount_in} {self.token_in} -> {self.amount_out} {self.token_out}\n' + '\n'.join(
            [str(s) for s in self.route])

    def __repr__(self) -> str:
        return str(self)



from dataclasses import dataclass, asdict
from decimal import Decimal
from typing import Union

from eth_typing import ChecksumAddress

from deanerfi.utils.w3 import cs_addr

TokenAmountType = Union[str, float, int, Decimal]


@dataclass(slots=True)
class ChainInfo:
    """Chain object stores the information needed to connect to an EVM blockchain with web3.py"""
    id: int
    name: str
    symbol: str
    rpc: str
    explorer: str = None
    scanner_api_key: str = None
    scanner_api_url: str = None
    inject_middleware: bool = False

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass(slots=True)
class ContractInfo:
    """Contract object stores the information needed to connect to a web3.Contract"""
    chain_id: int
    address: ChecksumAddress
    abi: list = None
    name: str = None

    def __post_init__(self) -> None:
        self.address = cs_addr(self.address)

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass(slots=True)
class TokenInfo:
    symbol: str
    name: str
    address: ChecksumAddress
    chain_id: int
    decimals: Decimal
    stable_coin: bool = False
    wl: bool = False

    def __post_init__(self) -> None:
        self.address = cs_addr(self.address)
        self.decimals = Decimal(self.decimals)

    @property
    def _base(self) -> Decimal:
        return Decimal('10') ** self.decimals

    def to_wei(self, amount: TokenAmountType) -> Decimal:
        """Returns the input amount of Ether in Wei format"""
        return Decimal(amount) * self._base

    def from_wei(self, amount: TokenAmountType) -> Decimal:
        """Returns the input amount of Wei in Ether format"""
        return Decimal(amount) / self._base

    def to_dict(self) -> dict:
        _dict = asdict(self)
        _dict['decimals'] = str(self.decimals)
        return _dict

    def __str__(self) -> str:
        return f'{self.name} ({self.symbol})'

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: "TokenInfo") -> bool:
        return (
            self.address == other.address and
            self.chain_id == other.chain_id
        )

    def __lt__(self, other: "TokenInfo") -> bool:
        return self.address < other.address


@dataclass(slots=True)
class Wallet:
    address: ChecksumAddress
    name: str

    def __post_init__(self) -> None:
        self.address = cs_addr(self.address)


class ChainListError(Exception):
    """Raised when there is an error with the provided config"""
    pass


class ChainList:

    def __init__(self, chains: [ChainInfo] = None) -> None:
        self._chains: dict[int, ChainInfo] = {}
        if chains:
            self.add_chains(chains)

    def add_chain(self, chain: ChainInfo) -> None:
        self._chains[chain.id] = chain
        setattr(self, chain.symbol, chain)

    def add_chains(self, chains: [ChainInfo]) -> None:
        for chain in chains:
            self.add_chain(chain)

    def __getitem__(self, chain_id: int) -> ChainInfo:
        try:
            _chain = self._chains[chain_id]
        except KeyError as e:
            raise ChainListError(f'Error: ChainInfo for chain id {chain_id} not configured')
        return _chain



from os import environ
from typing import Protocol

import requests

from deanerfi.w3 import ChainInfo
from deanerfi.utils.data import load_json, write_json
from deanerfi.utils.mongodb import MongoWrapper


class ChainRepo(Protocol):
    """Interface for a ChainRepo."""

    def get(self, chain_id: int) -> ChainInfo:
        """Get a ChainInfo by chain_id."""
        pass

    def save(self, chain_info: ChainInfo) -> None:
        """Save a ChainInfo to the repo."""
        pass


class LocalChainRepo:
    """In memory Chain Repo that can be loaded/ saved to a json file
    Default file location is '<LOCAL_REPO_PATH>/token-repo.json'
    """

    def __init__(self, file_path: str = None, save_to_file: bool = True) -> None:
        self.file_path = file_path
        self.save_to_file = save_to_file
        self._chains: dict[int, ChainInfo] = {}
        if self.file_path:
            self._load()

    def _load(self) -> None:
        """Load the repo from a json file."""
        chains_dict = load_json(self.file_path)
        for chain in chains_dict:
            if not chain.get('scanner_api_key'):
                chain['scanner_api_key'] = environ.get(f'{chain["symbol"]}_SCANNER_API_KEY')
            c_info = ChainInfo(**chain)
            self._chains[c_info.id] = c_info

    def get(self, chain_id: int) -> ChainInfo:
        """Get a ChainInfo by chain_id."""
        return self._chains.get(chain_id)

    def save(self, chain_info: ChainInfo) -> None:
        """Save a ChainInfo to the repo."""
        self._chains[chain_info.id] = chain_info
        if self.save_to_file and self.file_path:
            write_json(self.file_path, [chain.to_dict() for chain in self._chains.values()])


class MongoDBChainRepo:
    """Implements ChainRepo using MongoDB."""
    _table: str = 'chains'

    def __init__(self, mongo_wrapper: MongoWrapper = None) -> None:
        self._mongo = mongo_wrapper or MongoWrapper()

    def get(self, chain_id: int) -> ChainInfo:
        """Get a ChainInfo by chain_id."""
        response = self._mongo.get(self._table, {'id': chain_id})[0]
        _ = response.pop('_id')
        return ChainInfo(**response)

    def save(self, chain_info: ChainInfo) -> None:
        """Save a ChainInfo to the repo."""
        self._mongo.upsert(
            self._table,
            chain_info.to_dict(),
            {'id': chain_info.id}
        )


class HTTPDBChainRepo:
    """Implements ChainRepo using HTTP requests to an external API."""

    def __init__(self, api_url: str) -> None:
        self.api_url = api_url

    def get(self, chain_id: int) -> ChainInfo:
        response = requests.get(f'{self.api_url}/chains/{chain_id}')
        return ChainInfo(**response.json())

    def save(self, chain_info: ChainInfo) -> None:
        response = requests.post(f'{self.api_url}/chains/', json=chain_info.to_dict())



from ._adaptor import Web3Wrapper
from ._dtypes import (
    TokenInfo, Wallet, TokenAmountType, ChainInfo, ContractInfo
)
from ._repos import (
    TokenRepo, LocalTokenRepo, MongoDBTokenRepo,
    ChainRepo, LocalChainRepo, MongoDBChainRepo,
    ContractRepo, LocalContractRepo, MongoDBContractRepo
)
from . import api



from os import environ
from typing import Dict

from ._adaptors import DexPackage, DexRepo
from ._repo import LocalDexRepo
from ._uniswap import UNISWAP_V2_PACKAGE

local_repo_path = environ.get('LOCAL_REPO_PATH')

dex_repo_fp = f'{local_repo_path}/dex-repo.json' if local_repo_path else None

DEX_PKGS: Dict[str, DexPackage] = {
    UNISWAP_V2_PACKAGE.subtype: UNISWAP_V2_PACKAGE
}

DEX_REPO: DexRepo = LocalDexRepo(file_path=dex_repo_fp)



"""Top-level package for defi-tools."""

__author__ = """Dean Dominguez"""
__email__ = 'deanjdominguez@gmail.com'
__version__ = '0.1.0'

from os import environ

if environ.get('LOCAL', False):
    from dotenv import load_dotenv, find_dotenv
    load_dotenv(find_dotenv())



from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract
from web3.middleware import geth_poa_middleware

from ._dtypes import ChainInfo, TokenInfo
from deanerfi.utils.data import _load_pkg_data

ERC20_ABI = _load_pkg_data('erc20-abi.json')


class Web3Wrapper:
    """Implementation of Web3Handler abstraction for common web3 tasks"""

    def __init__(self, chain_id: int, rpc_url: str, inject_middleware: bool, name: str = None,
                 symbol: str = None) -> None:
        self.chain_id = chain_id
        self.name = name
        self.symbol = symbol
        self._w3 = Web3(
            Web3.HTTPProvider(rpc_url)
        )
        if inject_middleware:
            self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    def __enter__(self) -> 'Web3Wrapper':
        return self

    def __exit__(self, error: Exception, value: object, traceback: object) -> None:
        pass

    def get_contract(self, address: ChecksumAddress, abi: list) -> Contract:
        return self._w3.eth.contract(address=address, abi=abi)

    def get_token_info(self, address: ChecksumAddress) -> TokenInfo:
        token_contract = self.get_contract(address, ERC20_ABI)
        return TokenInfo(
            symbol=token_contract.functions.symbol().call(),
            name=token_contract.functions.name().call(),
            address=address,
            chain_id=self.chain_id,
            decimals=token_contract.functions.decimals().call(),
        )

    @staticmethod
    def from_chain_info(chain_info: ChainInfo) -> 'Web3Wrapper':
        return Web3Wrapper(
            chain_id=chain_info.id,
            rpc_url=chain_info.rpc,
            inject_middleware=chain_info.inject_middleware,
            name=chain_info.name,
            symbol=chain_info.symbol
        )



from typing import Protocol, Optional, List

from deanerfi.w3._dtypes import TokenInfo
from deanerfi.utils.data import load_json, _load_pkg_data, write_json, dataclass_from_dict
from deanerfi.utils.mongodb import MongoWrapper

ERC20_ABI = _load_pkg_data('erc20-abi.json')


class TokenRepo(Protocol):

    def get_chain(self, chain_id: int) -> [TokenInfo]:
        pass

    def get_token(self, chain_id: int, address: str) -> TokenInfo:
        pass

    def get_wl_token(self, chain_id: int) -> List[TokenInfo]:
        pass

    def save(self, token_info: TokenInfo) -> None:
        pass

    def whitelist(self, address: str, chain_id: int, wl: bool = True) -> None:
        pass

    def stable_coin(self, address: str, chain_id: int, stable_coin: bool = True) -> None:
        pass


class LocalTokenRepo:

    def __init__(self, file_path: str = None, save_to_file: bool = True) -> None:
        self.file_path = file_path
        self.save_to_file = save_to_file
        self._tokens: dict[int, dict[str, TokenInfo]] = {}
        if self.file_path:
            self._load()

    def _load(self) -> None:
        for token in load_json(self.file_path):
            self._save_token(TokenInfo(**token))

    def get_chain(self, chain_id: int) -> [TokenInfo]:
        chain_tokens = self._tokens.get(chain_id)
        if not chain_tokens:
            return []
        return list(chain_tokens.values())

    def get_token(self, chain_id: int, address: str) -> Optional[TokenInfo]:
        return self._tokens.get(chain_id, {}).get(address)

    def get_wl_token(self, chain_id: int) -> List[TokenInfo]:
        return [token for token in self.get_chain(chain_id=chain_id) if token.wl]

    def _save_token(self, token_info: TokenInfo) -> None:
        if token_info.chain_id not in self._tokens:
            self._tokens[token_info.chain_id] = {}
        self._tokens[token_info.chain_id][token_info.address] = token_info

    def _save_file(self) -> None:
        if self.file_path and self.save_to_file:
            tokens = []
            for chain_id, token_dict in self._tokens.items():
                tokens += [token.to_dict() for token in token_dict.values()]
            write_json(
                self.file_path,
                tokens
            )

    def save(self, token_info: TokenInfo) -> None:
        self._save_token(token_info)
        self._save_file()

    def whitelist(self, address: str, chain_id: int, wl: bool = True) -> None:
        token = self.get_token(chain_id=chain_id, address=address)
        if token:
            token.wl = wl
            self._save_file()

    def stable_coin(self, address: str, chain_id: int, stable_coin: bool = True) -> None:
        token = self.get_token(chain_id=chain_id, address=address)
        if token:
            token.stable_coin = stable_coin
            self._save_file()


class MongoDBTokenRepo:
    _table: str = 'tokens'

    def __init__(self) -> None:
        self._mongo = MongoWrapper()

    def get_chain(self, chain_id: int) -> [TokenInfo]:
        response = self._mongo.get(self._table, {'chain_id': chain_id})
        _tokens = []
        for t in response:
            _tokens.append(dataclass_from_dict(
                t,
                TokenInfo
            ))
        return _tokens

    def get_token(self, chain_id: int, address: str) -> TokenInfo:
        response = self._mongo.get_one(self._table, {'chain_id': chain_id, 'address': address})
        if response:
            return dataclass_from_dict(
                response,
                TokenInfo
            )

    def get_wl_token(self, chain_id: int) -> List[TokenInfo]:
        response = self._mongo.get(self._table, {'chain_id': chain_id, 'wl': True})
        _tokens = []
        for t in response:
            _ = t.pop('_id')
            _tokens.append(TokenInfo(**t))
        return _tokens

    def save(self, token_info: TokenInfo) -> None:
        self._mongo.upsert(
            table=self._table,
            data=token_info.to_dict(),
            id_dict={'chain_id': token_info.chain_id, 'address': token_info.address}
        )

    def whitelist(self, address: str, chain_id: int, wl: bool = True) -> None:
        token = self.get_token(chain_id=chain_id, address=address)
        if token:
            token.wl = wl
            self.save(token)

    def stable_coin(self, address: str, chain_id: int, stable_coin: bool = True) -> None:
        token = self.get_token(chain_id=chain_id, address=address)
        if token:
            token.stable_coin = stable_coin
            self.save(token)



from os import environ

from ._repos import (
    ContractRepo, LocalContractRepo, MongoDBContractRepo,
    ChainRepo, LocalChainRepo, MongoDBChainRepo,
    TokenRepo, LocalTokenRepo, MongoDBTokenRepo
)


local_repo_path = environ.get('LOCAL_REPO_PATH')

token_repo_fp = f'{local_repo_path}/token-repo.json' if local_repo_path else None
chain_repo_fp = f'{local_repo_path}/chain-repo.json' if local_repo_path else None

TOKEN_REPO: TokenRepo = LocalTokenRepo(file_path=token_repo_fp)
CHAIN_REPO: ChainRepo = LocalChainRepo(file_path=chain_repo_fp)



from decimal import Decimal
from typing import List, Tuple, Dict

from eth_typing import ChecksumAddress
from web3.contract import Contract
from deanerfi.common.log import api as log

from deanerfi.utils import graphql
from deanerfi.utils.data import _load_pkg_data
from ._adaptors import DexRepo, DexPoolAPI, DexPackage, DexAPI, ContractLoader, DexDataLoader
from ._dtypes import DexInfo, DexPoolInfo

UNISWAP_V2_DEFAULTS = _load_pkg_data('uniswap-v2-abis.json')


class UniswapV2PoolAPI(DexPoolAPI):

    def __init__(self, dex_id: str, address: ChecksumAddress, chain_id: int,
                 tokens: List[ChecksumAddress], pool_contract: Contract) -> None:
        super().__init__(dex_id=dex_id, chain_id=chain_id, address=address, tokens=tokens, pool_contract=pool_contract)

    def get_reserves(self) -> Dict[ChecksumAddress, Decimal]:
        res = self._contract.functions.getReserves().call()
        return {
            self.tokens[0]: Decimal(res[0]),
            self.tokens[1]: Decimal(res[1])
        }

    @staticmethod
    def build(config: DexPoolInfo, contract_loader: ContractLoader,
              pool_abi: list = None, **kwargs) -> 'UniswapV2PoolAPI':
        pool_abi = pool_abi or UNISWAP_V2_DEFAULTS['pool']
        return UniswapV2PoolAPI(
            dex_id=config.dex_id,
            address=config.address,
            chain_id=config.chain_id,
            tokens=config.tokens,
            pool_contract=contract_loader(config.chain_id, config.address, pool_abi)
        )


class UniswapV2API(DexAPI):

    def __init__(self, name: str, subtype: str, chain_id: int, trading_fee: Decimal, dex_id: str,
                 addresses: Dict[str, ChecksumAddress], contract_loader: ContractLoader,
                 pool_abi: list = None, router_abi: list = None, factory_abi: list = None) -> None:
        super().__init__(
            name=name, subtype=subtype, chain_id=chain_id, trading_fee=trading_fee,
            dex_id=dex_id, addresses=addresses, contract_loader=contract_loader
        )
        self._pool_abi = pool_abi or UNISWAP_V2_DEFAULTS['pool']
        router_abi = router_abi or UNISWAP_V2_DEFAULTS['router']
        factory_abi = factory_abi or UNISWAP_V2_DEFAULTS['factory']
        self.router: Contract = self._contract_loader(self.chain_id, addresses['router'], router_abi)
        self.factory: Contract = self._contract_loader(self.chain_id, addresses['factory'], factory_abi)

    def tradable(self, token0: ChecksumAddress, token1: ChecksumAddress) -> bool:
        raise NotImplementedError

    def quote(self, token0: ChecksumAddress, token1: ChecksumAddress, amount: int) -> Decimal:
        # resp = self.router.functions.getAmountsOut(amountIn=str(amount), path=[token0, token1]).call()
        resp = self.router.functions.getAmountsOut(amount, [token0, token1]).call()
        return Decimal(resp[-1])

    def get_pool_api(self, pool_info: DexPoolInfo) -> UniswapV2PoolAPI:
        pool_contract = self._contract_loader(self.chain_id, pool_info.address, self._pool_abi)
        return UniswapV2PoolAPI(
            dex_id=self.id,
            chain_id=self.chain_id,
            address=pool_info.address,
            tokens=pool_info.tokens,
            pool_contract=pool_contract
        )

    @staticmethod
    def build(config: DexInfo, contract_loader: ContractLoader, **kwargs) -> 'UniswapV2API':
        addresses = {
            'router': config.contracts['router'].address,
            'factory': config.contracts['factory'].address
        }
        return UniswapV2API(
            name=config.name, subtype=config.subtype, chain_id=config.chain_id, trading_fee=config.trading_fee,
            dex_id=config.id, addresses=addresses, contract_loader=contract_loader, pool_abi=config.pool_abi,
            router_abi=config.contracts['router'].abi, factory_abi=config.contracts['factory'].abi
        )


class UniswapV2DataLoader(DexDataLoader):
    BATCH_SIZE = 1000
    RESERVE_USD_GT = 100_000
    PAIRS_QUERY = """
    {{
      pairs(first: {batch_size}, skip: {skip_index}, where: {{reserveUSD_gt: {reserve_usd_gt}}}) {{
        id
        token0 {{
          id
        }}
        token1 {{
          id
        }}
      }}
    }}"""

    def _response_handler(self, data: dict, dex: str, chain_id: int) -> Tuple[List[DexPoolInfo], bool]:
        pools = data.get('data', {}).get('pairs')
        if not pools:
            return [], False
        pool_info = [
            DexPoolInfo(
                address=pool['id'],
                tokens=[pool['token0']['id'], pool['token1']['id']],
                dex_id=dex,
                chain_id=chain_id
            ) for pool in pools
        ]
        next_page = len(pools) == self.BATCH_SIZE
        return pool_info, next_page

    def load_data(self, dex_info: DexInfo, repo: DexRepo) -> None:
        if not dex_info.subgraph_url:
            log.info(__name__, f'No subgraph url for {dex_info.name} on chain {dex_info.chain_id}')
            return
        next_page = True
        skip_index = 0
        while next_page:
            query = self.PAIRS_QUERY.format(
                reserve_usd_gt=self.RESERVE_USD_GT,
                batch_size=self.BATCH_SIZE,
                skip_index=skip_index
            )
            data = graphql.query(dex_info.subgraph_url, query)
            pools, next_page = self._response_handler(data, dex=dex_info.id, chain_id=dex_info.chain_id)
            skip_index += self.BATCH_SIZE
            if pools:
                log.info(__name__, f'Saving {len(pools)} pools for {dex_info.name} on chain {dex_info.chain_id}')
                repo.save_pools(pools)


UNISWAP_V2_PACKAGE = DexPackage(
    subtype='UniswapV2',
    dex_api=UniswapV2API,
    pool_api=UniswapV2PoolAPI,
    data_loader=UniswapV2DataLoader
)



from typing import Protocol
import json
from typing import Optional

import requests
from web3 import Web3


class ChainScanner(Protocol):
    def get_tx(self, tx_hash):
        pass


class EtherScan:
    """Chain scanner for etherscan.com clones"""

    def __init__(self, url: str, api_key: str) -> None:
        self.base_url = url
        self.api_key = api_key

    def _make_req(self, params: dict) -> Optional[dict]:
        resp = requests.get(self.base_url, params=params)
        if resp.status_code != 200:
            return None
        return resp.json()

    def get_abi(self, addr: str) -> Optional[list]:
        params = {
            'module': 'contract',
            'action': 'getabi',
            'address': addr,
            'apikey': self.api_key
        }
        abi = self._make_req(params).get('result')
        if not abi:
            return None
        return json.loads(abi)

    def get_txs(self, addr: str) -> Optional[list]:
        params = {
            'module': 'account',
            'action': 'txlist',
            'address': addr,
            'startblock': 1,
            'endblock': 99999999,
            'sort': 'asc',
            'apikey': self.api_key
        }
        resp = self._make_req(params)
        if resp['message'] == 'OK':
            # return [EthScannerTx(**tx) for tx in resp['result']]
            for tx in resp['result']:
                tx['to'] = Web3.toChecksumAddress(tx['to'])
                tx['from'] = Web3.toChecksumAddress(tx['from'])
            return resp['result']
        else:
            return resp

    def get_internal_txs(self, tx_hash: str):
        params = {
            'module': 'account',
            'action': 'txlistinternal',
            'txhash': tx_hash,
            'apikey': self.api_key
        }
        resp = self._make_req(params)
        if resp['message'] == 'OK':
            return resp['result']
        else:
            return resp

    def get_token_txs(self, addr: str):
        params = {
            'module': 'account',
            'action': 'tokentx',
            'address': addr,
            'startblock': 1,
            'endblock': 99999999,
            'sort': 'asc',
            'apikey': self.api_key
        }
        resp = self._make_req(params)
        if resp['message'] == 'OK':
            return resp['result']
        else:
            return resp



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Type, Dict, List

from typing import Union, Callable, Optional
from deanerfi.common import log


@dataclass
class Event:
    name: str


@dataclass
class Command:
    name: str


Message = Union[Event, Command]

EventHandler = Callable[[Event], Optional[List[Message]]]
CommandHandler = Callable[[Command], Optional[List[Message]]]


class MessageBus(ABC):
    def __init__(self, event_handlers: Dict[Type[Event], List[EventHandler]] = None,
                 command_handlers: Dict[Type[Command], CommandHandler] = None, **kwargs) -> None:
        self.event_handlers = event_handlers
        self.command_handlers = command_handlers

    @abstractmethod
    def send_msg(self, msg: Message) -> None:
        raise NotImplementedError

    def send_msgs(self, msgs: List[Message]) -> None:
        raise NotImplementedError

    @abstractmethod
    def process_msgs(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def handle(self, msg: Message) -> None:
        raise NotImplementedError

    @abstractmethod
    def handle_event(self, event: Event) -> None:
        raise NotImplementedError

    @abstractmethod
    def handle_command(self, command: Command) -> None:
        raise NotImplementedError


class LocalMessageBus(MessageBus):

    def __init__(self, event_handlers: Dict[Type[Event], List[EventHandler]] = None,
                 command_handlers: Dict[Type[Command], CommandHandler] = None, **kwargs) -> None:
        super().__init__(event_handlers, command_handlers)
        self.msg_queue: List[Message] = []

    def send_msg(self, msg: Message) -> None:
        self.msg_queue.append(msg)

    def send_msgs(self, msgs: List[Message]) -> None:
        [self.msg_queue.append(msg) for msg in msgs]

    def process_msgs(self) -> None:
        while self.msg_queue:
            self.handle(self.msg_queue.pop(0))

    def handle(self, msg: Message) -> None:
        if isinstance(msg, Event):
            self.handle_event(msg)
        if isinstance(msg, Command):
            self.handle_command(msg)

    def handle_event(self, event: Event) -> None:
        log.api.info(__name__, 'handling event', data=event)

    def handle_command(self, command: Command) -> None:
        log.api.info(__name__, 'handling command', data=command)
        cmd_handler = self.command_handlers.get(type(command))
        if cmd_handler:
            return_msgs = cmd_handler(command)
            self.send_msgs(return_msgs)




from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from decimal import Decimal
from typing import List, Callable, Type, Dict

from eth_typing import ChecksumAddress
from web3.contract import Contract

from deanerfi.utils.data import dataclass_from_dict
from ._dtypes import DexInfo, DexPoolInfo
from deanerfi.w3 import ContractInfo

ContractLoader = Callable[[int, ChecksumAddress, list], Contract]


class DexAPI(ABC):
    """Interface for interacting with a DEX contract."""

    def __init__(self, name: str, subtype: str, chain_id: int, trading_fee: Decimal, dex_id: str,
                 addresses: Dict[str, ChecksumAddress], contract_loader: ContractLoader, **kwargs) -> None:
        self.name = name
        self.subtype = subtype
        self.chain_id = chain_id
        self.trading_fee = trading_fee
        self.id = dex_id
        self.chain_id = chain_id
        self.addresses = addresses
        self._contract_loader = contract_loader

    @abstractmethod
    def tradable(self, token0: ChecksumAddress, token1: ChecksumAddress) -> bool:
        """Check if a pair of tokens is tradable."""
        raise NotImplementedError

    @abstractmethod
    def quote(self, token0: ChecksumAddress, token1: ChecksumAddress, amount: int) -> Decimal:
        """Get the quote price for token0 in relation to token1."""
        raise NotImplementedError

    @abstractmethod
    def get_pool_api(self, pool_info: DexPoolInfo) -> 'DexPoolAPI':
        """Build a DEXPoolAPI from a DexPoolInfo object."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def build(config: DexInfo, contract_loader: ContractLoader, **kwargs) -> 'DexAPI':
        """Build a DEX contract from a DexInfo object."""
        raise NotImplementedError


class DexPoolAPI(ABC):
    """Interface for interacting with a DEX pool contract."""

    def __init__(self, dex_id: str, address: ChecksumAddress, chain_id: int,
                 tokens: List[ChecksumAddress], pool_contract: Contract) -> None:
        self.dex_id = dex_id
        self.chain_id = chain_id
        self.address = address
        self.tokens = tokens
        self._contract = pool_contract

    @abstractmethod
    def get_reserves(self) -> Dict[ChecksumAddress, Decimal]:
        """Get the current reserve of each token in the pool."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def build(config: DexPoolInfo, pool_abi: list, contract_loader: ContractLoader, **kwargs) -> 'DexPoolAPI':
        """Build a DexPoolAPI from a DexPoolInfo."""
        raise NotImplementedError

    def __eq__(self, other: "DexPoolAPI") -> bool:
        return self.tokens == other.tokens

    def __str__(self) -> str:
        return f'{self.dex_id}: ' + ' - '.join([str(t) for t in self.tokens])

    def __repr__(self) -> str:
        return str(self)


class DexRepo(ABC):
    """Interface for interacting with a DEX repository."""

    @staticmethod
    def _pool_from_dict(pool_dict: dict) -> DexPoolInfo:
        return dataclass_from_dict(
            pool_dict,
            DexPoolInfo
        )

    @staticmethod
    def _config_from_dict(dex_dict: dict) -> DexInfo:
        """Convert a DEX config dict to a DEXInfo object."""
        for contract_name, contract_dict in dex_dict['contracts'].items():
            dex_dict['contracts'][contract_name] = ContractInfo(**contract_dict)
        return dataclass_from_dict(
            dex_dict,
            DexInfo
        )

    @abstractmethod
    def get_configs(self, subtype: str = None, chain_id: int = None) -> List[DexInfo]:
        """Get DexInfo objects for all DEXes of a given subtype on a given chain."""
        raise NotImplementedError

    @abstractmethod
    def get_config(self, dex_id: str = None) -> DexInfo:
        """Get DexInfo object for a given DEX."""
        raise NotImplementedError

    @abstractmethod
    def delete_dex(self, dex_id: str) -> None:
        """Delete a DEX from the repository."""
        raise NotImplementedError

    @abstractmethod
    def token_pools(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        """Returns the Dex's pools containing only the supplied tokens"""
        raise NotImplementedError

    @abstractmethod
    def pools_w_tokens(self, chain_id: int, tokens: List[ChecksumAddress]) -> List[DexPoolInfo]:
        """Returns all pools that contain any of the supplied tokens"""
        raise NotImplementedError

    @abstractmethod
    def dex_pools(self, dex_id: str) -> List[DexPoolInfo]:
        """Gets all pools for the given dex"""
        raise NotImplementedError

    @abstractmethod
    def get_pool_info(self, chain_id: int, address: ChecksumAddress) -> DexPoolInfo:
        """Gets all pools for the given dex"""
        raise NotImplementedError

    @abstractmethod
    def chain_pools(self, chain_id: int) -> List[DexPoolInfo]:
        """Gets all pools for the given dex"""
        raise NotImplementedError

    @abstractmethod
    def save_pools(self, pools: List[DexPoolInfo]) -> None:
        """Saves the pools to persistent storage"""
        raise NotImplementedError

    @abstractmethod
    def save_configs(self, configs: List[DexInfo]) -> None:
        """Saves the DexInfo to persistent storage"""
        raise NotImplementedError


class DexDataLoader(ABC):
    """Interface for loading Dex data into the DexRepo. Most data is loaded from The Graph."""

    @abstractmethod
    def load_data(self, dex_info: DexInfo, repo: DexRepo) -> None:
        raise NotImplementedError


@dataclass
class DexPackage:
    """Class that groups the dex config (DexInfo), api (DexAPI), and data loader (DexDataLoader) into a package that can
    keep track of the components required to fully integrate with a dex.
    """
    subtype: str
    dex_api: Type[DexAPI]
    pool_api: Type[DexPoolAPI]
    data_loader: Type[DexDataLoader] = field(default=None)



from eth_typing import ChecksumAddress
from web3 import Web3


def cs_addr(address: str) -> ChecksumAddress:
    """Simple wrapper function that takes a string and returns a ChecksumAddress"""
    return Web3.toChecksumAddress(address)




from typing import Protocol, List
from deanerfi.dex import SwapRoute, DexPoolInfo, Swap
from eth_typing import ChecksumAddress
import networkx as nx

from deanerfi.utils.w3 import cs_addr


class TradeRouter(Protocol):

    def route(self, token0: ChecksumAddress, token1: ChecksumAddress, max_swaps: int = 2) -> List[SwapRoute]:
        """Returns all possible swap routes. Routes initialized with 0 tokens traded"""
        pass


class TokenGraphTradeRouter:
    """Trades Router that utilizes a graph with Tokens as nodes and pools as edges"""

    def __init__(self, pools: List[DexPoolInfo] = None):
        self.graph = nx.MultiGraph()
        if pools:
            self.add_pools(pools)

    def route(self, token0: ChecksumAddress, token1: ChecksumAddress, max_swaps: int = 2) -> List[SwapRoute]:
        paths = nx.all_simple_edge_paths(self.graph, source=token0, target=token1, cutoff=max_swaps)
        if not paths:
            return []
        return [self.parse_path(path) for path in paths]

    def loop_route(self, token0: ChecksumAddress, max_swaps: int = 2) -> List[SwapRoute]:
        start_paths = self.route(token0=token0, token1=self.graph.neighbors(token0), max_swaps=max_swaps - 1)
        paths = []
        for path in start_paths:
            ret_paths = self.route(token0=path.token_out, token1=path.token_in, max_swaps=1)
            for ret_path in ret_paths:
                # if path.route[0] != ret_path.route[0]:
                paths.append(
                    SwapRoute(route=path.route + ret_path.route)
                )
        return paths

    @staticmethod
    def parse_path(path: List[ChecksumAddress]) -> SwapRoute:
        swap_route = SwapRoute()
        for swap in path:
            swap_route.add_swap(
                Swap(
                    token_in=cs_addr(swap[0]),
                    token_out=cs_addr(swap[1]),
                    pool_address=cs_addr(swap[2]),
                )
            )
        return swap_route

    def add_node(self, token_addr: ChecksumAddress) -> None:
        if not self.graph.has_node(token_addr):
            self.graph.add_node(token_addr)

    def add_nodes(self, token_addrs: List[ChecksumAddress]) -> None:
        for token_addr in token_addrs:
            self.add_node(token_addr)

    def add_pool(self, pool: DexPoolInfo) -> None:
        self.add_nodes(pool.tokens)
        self.graph.add_edge(pool.tokens[0], pool.tokens[1], pool.address)

    def add_pools(self, pools: List[DexPoolInfo]) -> None:
        for pool in pools:
            self.add_pool(pool)



from typing import Protocol, List
from deanerfi.w3 import TokenInfo
from deanerfi.dex import DexPoolAPI


class Arbitrageur(Protocol):

    def run(self) -> None:
        pass


class PoolArbitrageur:

    def __init__(self, tokens: List[TokenInfo], pools: List[DexPoolAPI]) -> None:
        self.tokens = tokens
        self.tokens.sort()
        self.pools = pools

    def run(self) -> None:
        print(f'{self.tokens[0].name} - {self.tokens[0].address}')
        print(f'{self.tokens[1].name} - {self.tokens[1].address}')
        for pool in self.pools:
            print(pool.dex_id)
            pool_reserves = pool.get_reserves()
            num_t0 = self.tokens[0].from_wei(pool_reserves[self.tokens[0].address])
            num_t1 = self.tokens[1].from_wei(pool_reserves[self.tokens[1].address])

            print(f'  {self.tokens[0].symbol}: {int(num_t0)} ')
            print(f'  {self.tokens[1].symbol}: {int(num_t1)}')

            print(f'  {self.tokens[0].symbol}/{self.tokens[1].symbol}: {num_t0/num_t1}')
            print(f'  {self.tokens[1].symbol}/{self.tokens[0].symbol}: {num_t1 / num_t0}')








from typing import Callable

import pika
from os import environ
from abc import ABC


class RabbitMQWrapper(ABC):
    _connection = None
    _channel = None
    exchange = None

    def __init__(self, host: str = None, port: int = None, user: str = None, password: str = None) -> None:
        self._host = host or environ.get('RABBITMQ_HOST')
        self._port = port or environ.get('RABBITMQ_PORT')
        self._username = user or environ.get('RABBITMQ_USER')
        self._password = password or environ.get('RABBITMQ_PASSWORD')
        self.connect()

    def connect(self) -> None:
        self._connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=self._host,
                port=self._port,
                credentials=pika.PlainCredentials(
                    username=self._username,
                    password=self._password
                )
            )
        )
        self._channel = self._connection.channel()


class RabbitMQPub(RabbitMQWrapper):

    def __init__(self, host: str = None, port: int = None, user: str = None, password: str = None) -> None:
        super().__init__(host=host, port=port, user=user, password=password)

    def set_exchange(self, name: str, ex_type: str = 'direct') -> None:
        self.exchange = name
        self._channel.exchange_declare(
            exchange=name, exchange_type=ex_type
        )

    def publish(self, msg, routing_key) -> None:
        self._channel.basic_publish(
            exchange=self.exchange,
            routing_key=routing_key,
            body=msg
        )


def callback(ch, method, properties, body):
    print(f'ch: {ch}')
    print(f'method: {method}')
    print(f'properties: {properties}')
    print(f'body: {body}')


class RabbitMQSub(RabbitMQWrapper):
    _callback: Callable = callback
    _queue = None

    def __init__(self, host: str = None, port: int = None, user: str = None, password: str = None) -> None:
        super().__init__(host=host, port=port, user=user, password=password)

    def set_callback(self, func: Callable) -> None:
        self._callback = func

    def set_queue(self, exchange_name: str, queue_name: str, routing_key: str):
        self._queue = self._channel.queue_declare(queue_name)
        self._channel.queue_bind(
            exchange=exchange_name,
            queue=queue_name,
            routing_key=routing_key
        )

    def run(self):
        self._channel.basic_consume(
            on_message_callback=self._callback,
            queue=self._queue.method.queue
        )



from typing import List
from ._dtypes import DexInfo, DexPoolInfo, SwapRoute, Swap
from ._adaptors import DexAPI, DexPoolAPI
from ._repo import DexRepo
from . import _config
from deanerfi.w3.api import get_contract, get_token
from web3.exceptions import BadFunctionCallOutput
from deanerfi.common.log import api as log


def get_dex_repo() -> DexRepo:
    return _config.DEX_REPO


def set_dex_repo(dex_repo: DexRepo) -> None:
    _config.DEX_REPO = dex_repo


def get_dex_configs(chain_id: int = None, subtype: str = None) -> List[DexInfo]:
    """Returns the DexInfos that are configured in the DexRepo"""
    return _config.DEX_REPO.get_configs(chain_id=chain_id, subtype=subtype)


def build_dex(dex_config: DexInfo) -> DexAPI:
    """Builder function that takes the DexInfo and returns the DexAPI"""
    return _config.DEX_PKGS[dex_config.subtype].dex_api.build(config=dex_config, contract_loader=get_contract)


def update_dex_pools(dex_config: DexInfo) -> None:
    """Updates the available pools for the given dex in the DexRepo"""
    data_loader = _config.DEX_PKGS[dex_config.subtype].data_loader()
    log.info(__name__, f'Updating Pools for {dex_config.name} on Chain #{dex_config.chain_id}')
    data_loader.load_data(dex_info=dex_config, repo=get_dex_repo())


def update_pools() -> None:
    """Updates all known dexs information in the DexRepo"""
    for dex_config in get_dex_configs():
        update_dex_pools(dex_config=dex_config)


def get_dex_config(dex_id: str) -> DexInfo:
    return _config.DEX_REPO.get_config(dex_id=dex_id)


def get_pool_api(pool_info: DexPoolInfo) -> DexPoolAPI:
    dex_config = get_dex_config(pool_info.dex_id)
    return _config.DEX_PKGS[dex_config.subtype].pool_api.build(
        config=pool_info,
        pool_abi=dex_config.pool_abi,
        contract_loader=get_contract
    )


def print_swap_route(swap_route: SwapRoute, chain_id: int) -> None:
    dex_repo = get_dex_repo()
    token_in = get_token(chain_id=chain_id, address=swap_route.token_in)
    token_out = get_token(chain_id=chain_id, address=swap_route.token_out)
    print(
        f'Swap Route: {token_in.from_wei(swap_route.amount_in)} {token_in.symbol} -> {token_out.from_wei(swap_route.amount_out)} {token_out.symbol}')
    try:
        for swap in swap_route.route:
            pool_info = dex_repo.get_pool_info(chain_id=chain_id, address=swap.pool_address)
            token_in = get_token(chain_id=chain_id, address=swap.token_in)
            token_out = get_token(chain_id=chain_id, address=swap.token_out)
            print(f' - {pool_info.dex_id}: {swap.amount_in} {token_in.symbol} -> {swap.amount_out} {token_out.symbol}')
    except BadFunctionCallOutput:
        log.error(__name__, f'BadFunctionCallOutput:', data=swap_route)


def quote_swap(swap: Swap, chain_id: int) -> None:
    pool_info = _config.DEX_REPO.get_pool_info(chain_id=chain_id, address=swap.pool_address)
    dex_api = build_dex(get_dex_config(pool_info.dex_id))
    swap.amount_out = dex_api.quote(token0=swap.token_in, token1=swap.token_out, amount=int(swap.amount_in))


def quote_route(swap_route: SwapRoute, chain_id: int) -> None:
    len_swaps = len(swap_route.route) - 1
    for i, swap in enumerate(swap_route.route):
        quote_swap(swap, chain_id)
        if i != len_swaps:
            swap_route.route[i + 1].amount_in = swap_route.route[i].amount_out



from ._adaptors import DexAPI, DexPoolAPI, DexRepo, DexDataLoader, DexPackage
from ._repo import LocalDexRepo, MongoDBDexRepo
# from ._dtypes import DexInfo, DexPoolInfo, UniswapV2DexInfo, Swap, SwapRoute
from ._dtypes import DexInfo, DexPoolInfo, Swap, SwapRoute

from . import api



from typing import List

from web3.contract import Contract

from . import _config
from ._adaptor import Web3Wrapper
from ._dtypes import TokenInfo, ChainInfo
from ._repos import ChainRepo, TokenRepo
from deanerfi.utils.w3 import cs_addr


def get_chain_repo() -> ChainRepo:
    return _config.CHAIN_REPO


def set_chain_repo(chain_repo: ChainRepo) -> None:
    _config.CHAIN_REPO = chain_repo


def get_token_repo() -> TokenRepo:
    return _config.TOKEN_REPO


def set_token_repo(token_repo: TokenRepo) -> None:
    _config.TOKEN_REPO = token_repo


def get_chain_info(chain_id: int) -> ChainInfo:
    chain_repo = get_chain_repo()
    return chain_repo.get(chain_id=chain_id)


def get_token(address: str, chain_id: int, save: bool = True) -> TokenInfo:
    address = cs_addr(address)
    token_repo = get_token_repo()
    token = token_repo.get_token(chain_id, address)
    if token:
        # CONFIG.LOGGER.info(__name__, 'TokenInfo found in TokenRepo', _token.to_dict())
        return token
    chain_info = get_chain_info(chain_id=chain_id)
    # CONFIG.LOGGER.info(__name__, 'ChainInfo: ', _chain)
    with Web3Wrapper.from_chain_info(chain_info) as w3_uow:
        token = w3_uow.get_token_info(address)
        # CONFIG.LOGGER.info(__name__, 'TokenInfo not found in TokenRepo. Loaded from contract', _token.to_dict())
    if save:
        token_repo.save(token)
    return token


def get_contract(chain_id: int, address: str, abi: list) -> Contract:
    chain_info = get_chain_info(chain_id=chain_id)
    address = cs_addr(address)
    with Web3Wrapper.from_chain_info(chain_info) as w3_uow:
        contract = w3_uow.get_contract(address=address, abi=abi)
    return contract


def get_wl_tokens(chain_id: int) -> List[TokenInfo]:
    return _config.TOKEN_REPO.get_wl_token(chain_id=chain_id)
